#include <string>
#include <vector>
#include <iostream>
#include <algorithm>
#include <cstdio>

static bool maysort = true;

/* Characters can never be split. */
typedef std::string Character;
typedef std::vector<Character> Item;
typedef std::vector<Item> Itemlist;

static const Item Import(const std::string& s)
{
	Item result(s.size());
	for(unsigned a=0; a<s.size(); ++a)
	{
		Character ch;
		switch(s[a])
		{
			case ' ':
			case '\'': case '"':
			case ',': case '$':
			case '{': case '}':
			case '&': case '|':
			case '(': case ')':
			case '\\': ch += '\\'; //passthru
			default: ch += s[a];
		}
		result[a] = ch;
	}
	return result;
}
static const std::string Export(const Item& item)
{
	switch(item.size())
	{
		case 0: return "";
		case 1: return item[0];
		default:
			std::string result;
			for(unsigned a=0; a<item.size(); ++a) result += item[a];
			return result;
	}
}

static const Item Characterify(const Itemlist& items)
{
	std::string token = "{";
	for(unsigned a=0; a<items.size(); ++a)
	{
		if(a > 0) token += ',';
		token += Export(items[a]);
	}
	token += "}";

	Item result;
	result.push_back(token);
	return result;
}

static const Item Slice(const Item& item, unsigned pos, unsigned length)
{
	return Item(item.begin() + pos,
				item.begin() + pos+length);
}
static const Itemlist Pack(Itemlist lines)
{
Restart:
	unsigned max_length = 0;
	for(unsigned a=0; a<lines.size(); ++a)
		max_length = std::max(max_length, (unsigned)lines[a].size());
	
	unsigned best_winning = 0;
	unsigned best_begin   = 0;
	unsigned best_end     = 0;
	
	/*
	fprintf(stderr, "Using %lu bytes of temporary space\n",
		(unsigned long)(max_length * lines.size() * lines.size()));
	*/
	
	std::vector<char> match_map((max_length+1) * lines.size() * lines.size(), 0);
	
	for(unsigned len=0; len<=max_length; ++len)
	{
		for(unsigned a=0; a<lines.size(); ++a)
		{
			if(lines[a].size() < len) continue;
			for(unsigned b=a+1; b<lines.size(); ++b)
			{
				if(lines[b].size() < len) continue;

				match_map[(a + b*lines.size())*(max_length+1) + len]
				|= 1*std::equal(lines[a].begin(), lines[a].begin() + len, lines[b].begin());
				
				match_map[(a + b*lines.size())*(max_length+1) + len]
				|= 2*std::equal(lines[a].begin() + lines[a].size() - len, lines[a].end(),
				                lines[b].begin() + lines[b].size() - len);
			}
		}
	}

	for(unsigned phase=1; phase<=2; ++phase)
	{
		/* phase 1: Find best winning
		 * phase 2: Find the same "best winning", and apply it
		 */
		
		const unsigned min_begin = phase == 1 ? 0            : best_begin;
		const unsigned max_begin = phase == 1 ? max_length+1 : best_begin+1;
		
		for(unsigned begin=min_begin; begin<max_begin; ++begin)
		{
			const unsigned min_end = phase == 1 ? !begin             : best_end;
			const unsigned max_end = phase == 1 ? max_length-begin+1 : best_end+1;
			
			for(unsigned end=min_end; end < max_end; ++end)
			{
				std::vector<bool> ignore(lines.size(), false);
				
				for(unsigned a=0; a<lines.size(); ++a)
				{
					if(ignore[a]) continue;
					if(lines[a].size() < begin+end) continue;
					
					const unsigned aendpos = lines[a].size() - end;
					
					unsigned num_matches = 0;

					std::vector<unsigned> matchlist;
					
					for(unsigned b=a+1; b<lines.size(); ++b)
					{
						if(lines[b].size() >= begin+end)
						{
							if((match_map[(a + b*lines.size())*(max_length+1) + begin]&1)
							&& (match_map[(a + b*lines.size())*(max_length+1) + end]&2))
							{
								// match.
								if(phase == 2) matchlist.push_back(b);
								++num_matches;
								continue;
							}
						}
						// if maysort==false, only adjacent lines
						// can be candidates for joining.
						if(!maysort) break;
					}
					
					if(num_matches == 0) continue;
					
					Item beginpart = Slice(lines[a], 0,       begin);
					Item endpart   = Slice(lines[a], aendpos, end);
					
					unsigned winning_per_one
						= (Export(beginpart).size() + Export(endpart).size()) * max_length
						+ std::max(begin, end);
					
					//if(winning_per_one <= 2*max_length+2) continue;
					
					unsigned winning = winning_per_one * (num_matches+1);
					
					switch(phase)
					{
						case 1:
							if(winning > best_winning)
							{
								best_winning = winning;
								best_begin   = begin;
								best_end	 = end;
							}
							break;
						case 2:
							if(winning == best_winning)
							{
								/* Found the right combination */

								Itemlist slices;
								
								slices.push_back(Slice(lines[a], begin, lines[a].size()-end-begin));
								
								// Erase the slices, except "a" */
								
								for(unsigned c=0; c<matchlist.size(); ++c)
								{
									unsigned b = matchlist[c];
	   								slices.push_back(Slice(lines[b], begin, lines[b].size()-end-begin));
								}
								for(unsigned c=matchlist.size(); c-- > 0; )
									lines.erase(lines.begin() + matchlist[c]);
								
								lines[a].clear();
								lines[a].insert(lines[a].end(), beginpart.begin(), beginpart.end());

								slices = Pack(slices);
								Item middlepart = Characterify(slices);

								lines[a].insert(lines[a].end(), middlepart.begin(),middlepart.end());
								lines[a].insert(lines[a].end(), endpart.begin(),   endpart.end());

								goto Restart;
							}
							break;
					}
					
					for(unsigned b=0; b<matchlist.size(); ++b)
						ignore[matchlist[b]] = true;
				}
			}
		}
		
		if(best_winning <= 0) break;
	}
	return lines;
}


//#define _GNU_SOURCE
#include <unistd.h>
#include <getopt.h>

int main(int argc,char**argv)
{
	for(;;)
	{
		int option_index = 0;
		static struct option long_options[] =
		{
			{"help",	 0, 0,'h'},
			{"version",  0, 0,'V'},
			{"nosort",   0, 0,'n'},
			{0,0,0,0}
		};
		int c = getopt_long(argc, argv, "hVn", long_options, &option_index);
		if(c==-1) break;
		switch(c)
		{
			case 'V':
			{
				std::cout << VERSION"\n";
				return 0;
			}
			case 'h':
			{
				std::cout <<
					"shellpack v"VERSION" - Copyright (C) 1992,2006 Bisqwit (http://iki.fi/bisqwit/)\n"
					"\n"
					"Usage: shellpack [<options>] < wordlist\n"
					"\n"
					"Transforms a wordlist into a compressed expression that can be used\n"
					"in a sh-compatible shell command line.\n"
					"\n"
					" --help, -h     This help\n"
					" --version, -V  Displays version information\n"
					" --nosort, -n   Retains the original order of words (faster, but less powerful)\n"
					"\n";
				return 0;
			}
			case 'n':
				maysort = false;
				break;
			default: goto argerror;
		}
	}
	
	if(argc != optind)
	{
	argerror:
		std::cerr << "shellpack: Invalid parameters. Try `shellpack --help'\n";
		return 1;
	}

	Itemlist lines;

	for(;;) 
	{
		std::string s;
		std::getline(std::cin, s);
		if(!std::cin.good())break;
		lines.push_back(Import(s));
	}
	if(maysort)
		std::sort(lines.begin(), lines.end());

	/*Pack(lines);Pack(lines);Pack(lines);Pack(lines);Pack(lines);Pack(lines);Pack(lines);
	Pack(lines);Pack(lines);Pack(lines);*/
	/*Pack(lines);Pack(lines);Pack(lines);Pack(lines);*/
	lines = Pack(lines);
	
	for(unsigned a=0; a<lines.size(); ++a)
	{
		if(a>0) std::printf(" ");
		std::printf("%s", Export(lines[a]).c_str());
	}
	std::printf("\n");
	
	return 0;
}
